using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.IO;
using System.Drawing.Imaging;

namespace Gpm2Utility {
  public class Edy {
    private Color [] _colors = new Color [256];
    private byte [] _buffer;
    private int _width;
    private int _height;
    private Bitmap _bitmap;
    private double _entropy;

    public Edy (string fileName) {
      byte [] pal = Properties.Resources.Gpm2_pal;
      for (int i = 0; i < 256; i++) {
        int j = 24 + i * 4;
        //Color color = Color.FromArgb (pal [j + 3], pal [j + 0], pal [j + 1], pal [j + 2]);
        Color color = Color.FromArgb (pal [j + 0], pal [j + 1], pal [j + 2]);
        _colors [i] = color;
      }
      //_colors [0] = Color.Black;
      //for (int i = 0; i < 256; i++) {
      //  Console.WriteLine ("R:{0} G:{1} B:{2} (A:{3})", _colors [i].R, _colors [i].G, _colors [i].B, _colors [i].A);
      //}

      Stream stream = new FileStream (fileName, System.IO.FileMode.Open);
      _buffer = new byte [stream.Length];
      stream.Read (_buffer, 0, (int) stream.Length);
      stream.Close ();
    }

    public int EstimateWidth () {
      double bestEntropy = -1;
      int bestWidth = -1;

      for (int width = 8; width < Math.Sqrt (_buffer.Length) * 2; width++) {
        int height = _buffer.Length / width;
        if ((_buffer.Length % width) != 0) {
          continue;
        }

        double entropy = ComputeEntropy (width, height);

        if ((bestEntropy == -1) || (entropy < bestEntropy)) {
          bestWidth = width;
          bestEntropy = entropy;
        }
      }
      return (bestWidth);
    }

    private double ComputeEntropy (int width, int height) {
      double entropy = 0;
      for (int y = 0; y < height; y++) {
        for (int x = 0; x < width; x++) {
          if ((x + y * width) < _buffer.Length) {
            entropy += ComputeEntropy (x, y, width, height);
          }
        }
      }
      entropy /= _buffer.Length;
      return (entropy);
    }

    private double ComputeEntropy (int x, int y, int width, int height) {
      double entropy = 0;
      int nb = 0;
      int pow = 1;
      Color centerColor = _colors [_buffer [x + y * width]];
      if ((x - 1) >= 0) {
        entropy += Math.Pow (ComputeDelta (centerColor, _colors [_buffer [x - 1 + y * width]]), pow);
        nb++;
      }
      if ((x + 1) < width) {
        if ((x + 1 + y * width) < _buffer.Length) {
          entropy += Math.Pow (ComputeDelta (centerColor, _colors [_buffer [x + 1 + y * width]]), pow);
          nb++;
        }
      }
      if ((y - 1) >= 0) {
        entropy += Math.Pow (ComputeDelta (centerColor, _colors [_buffer [x + (y - 1) * width]]), pow);
        nb++;
      }
      if ((y + 1) < height) {
        if ((x + (y + 1) * width) < _buffer.Length) {
          entropy += Math.Pow (ComputeDelta (centerColor, _colors [_buffer [x + (y + 1) * width]]), pow);
          nb++;
        }
      }
      //entropy /= nb;
      return (entropy);
    }

    private double ComputeDelta (Color color1, Color color2) {
      return ((Math.Abs (color1.R - color2.R) + Math.Abs (color1.G - color2.G) + Math.Abs (color1.B - color2.B)) / 3);
    }

    //public void SetWidth (int width) {
    //  _width = width;
    //  int height = _buffer.Length / _width;
    //  if ((_buffer.Length % _width) != 0) {
    //    height++;
    //  }
    //  _height = height;
    //  CreateBitmap ();
    //}

    public unsafe void CreateBitmap (int width) {
      _width = width;
      int height = _buffer.Length / _width;
      if ((_buffer.Length % _width) != 0) {
        height++;
      }
      _height = height;
      _entropy = ComputeEntropy (_width, _height);

      _bitmap = new Bitmap (_width, _height, PixelFormat.Format8bppIndexed);
      //_bitmap = new Bitmap (_width, _height, PixelFormat.Format32bppPArgb);
      ColorPalette colorPalette = _bitmap.Palette;
      for (int i = 0; i < 256; i++) {
        colorPalette.Entries [i] = _colors [i];
      }
      _bitmap.Palette = colorPalette;

      BitmapData bmd = _bitmap.LockBits (new Rectangle (0, 0, _bitmap.Width, _bitmap.Height), ImageLockMode.WriteOnly, _bitmap.PixelFormat);
      int index = 0;
      for (int y = 0; y < _height; y++) {
        for (int x = 0; x < _width; x++) {
          if (index < _buffer.Length) {
            //SetPixel (x, y, _buffer [index]);

            byte* p = (byte*) bmd.Scan0.ToPointer ();
            int offset = y * bmd.Stride + (x);
            p [offset] = _buffer [index];
            //p [offset] = 0;

            //_bitmap.SetPixel (x, y, _colors [_buffer [index]]);
            index++;
          }
        }
      }



      _bitmap.UnlockBits (bmd);
    }

    public unsafe void SetPixel (int x, int y, byte c) {
    }

    public double Entropy {
      get {
        return (_entropy);
      }
    }

    public int Size {
      get {
        return (_buffer.Length);
      }
    }

    public Bitmap Bitmap {
      get {
        return (_bitmap);
      }
    }
  }
}
